# Copyright 2015 Twitter, Inc and other contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from functools import wraps
import concurrent.futures
import sys
import threading
from unishark.util import exc_info_to_string
from unishark.runner import out
from unishark.exception import MultipleErrors


def data_driven(*list_of_dicts, **dict_of_lists):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if list_of_dicts:
                for params in list_of_dicts:
                    if type(params) is not dict:
                        raise TypeError('%r is not a dict.' % params)
                    kwargs.update(params)
                    func(*args, **kwargs)
                return
            if dict_of_lists:
                values_list = dict_of_lists.values()
                for values in values_list:
                    if type(values) is not list:
                        raise TypeError('%r is not a list.' % values)
                keys = dict_of_lists.keys()
                for values in zip(*values_list):
                    params = dict(zip(keys, values))
                    kwargs.update(params)
                    func(*args, **kwargs)
                return
        return wrapper
    return decorator


def multi_threading_data_driven(max_workers, *list_of_dicts, **dict_of_lists):
    if type(max_workers) is not int:
        raise TypeError('max_workers is not an integer.')
    if max_workers <= 0:
        raise ValueError('max_workers <= 0.')

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if list_of_dicts:
                for params in list_of_dicts:
                    if type(params) is not dict:
                        raise TypeError('%r is not a dict.' % params)
                with concurrent.futures.ThreadPoolExecutor(max_workers) as executor:
                    futures = []
                    for params in list_of_dicts:
                        kwargs.update(params)
                        futures.append(executor.submit(_fn_with_traceback, func, *args, **kwargs))
                    _handle_futures(futures)
                return
            if dict_of_lists:
                values_list = dict_of_lists.values()
                for values in values_list:
                    if type(values) is not list:
                        raise TypeError('%r is not a list.' % values)
                values_groups = list(zip(*values_list))
                with concurrent.futures.ThreadPoolExecutor(max_workers) as executor:
                    futures = []
                    keys = dict_of_lists.keys()
                    for values in values_groups:
                        params = dict(zip(keys, values))
                        kwargs.update(params)
                        futures.append(executor.submit(_fn_with_traceback, func, *args, **kwargs))
                    _handle_futures(futures)
                return
        return wrapper
    return decorator


def _fn_with_traceback(fn, *args, **kwargs):
    try:
        fn(*args, **kwargs)
        return threading.current_thread().ident, None
    except Exception:
        return threading.current_thread().ident, exc_info_to_string(sys.exc_info())


def _handle_futures(futures):
    results = [f.result() for f in concurrent.futures.as_completed(futures)]
    idents = [r[0] for r in results]
    for ident in idents:
        if ident in out.buff_dict:
            buff = out.buff_dict.pop(ident)
            out.write(buff.getvalue())
            buff.seek(0)
            buff.truncate()
            out.buff_queue.append(buff)
    exc_info_strs = [r[1] for r in results if r[1] is not None]
    if exc_info_strs:
        raise MultipleErrors(exc_info_strs)


# Copyright 2015 Twitter, Inc and other contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import time
import sys
import traceback
from unishark.util import (get_long_class_name, get_long_method_name, get_module_name)
import threading
from collections import deque
from inspect import ismodule

_io = None
if sys.version_info[0] < 3:  # python 2.7
    _io = __import__('StringIO')
else:
    _io = __import__('io')

if _io is None or not ismodule(_io):
    raise ImportError


def _make_buffer():
    return _io.StringIO()


class _PooledIOBuffer(object):
    _lock = threading.RLock()

    def __init__(self):
        self.buff_queue = deque()
        self.buff_queue.append(_make_buffer())
        self.buff_dict = dict()

    def _get_buff(self):
        with _PooledIOBuffer._lock:
            if not self.buff_queue:
                return _make_buffer()
            else:
                return self.buff_queue.popleft()

    def write(self, *args, **kwargs):
        i = threading.current_thread().ident
        if i not in self.buff_dict:
            buff = self._get_buff()
            self.buff_dict[i] = buff
        self.buff_dict[i].write(*args, **kwargs)

    def getvalue(self, *args, **kwargs):
        i = threading.current_thread().ident
        return self.buff_dict[i].getvalue(*args, **kwargs) if i in self.buff_dict else None

    def flush(self, *args, **kwargs):
        i = threading.current_thread().ident
        if i in self.buff_dict:
            self.buff_dict[i].flush(*args, **kwargs)

    def seek(self, *args, **kwargs):
        i = threading.current_thread().ident
        if i in self.buff_dict:
            self.buff_dict[i].seek(*args, **kwargs)

    def truncate(self, *args, **kwargs):
        i = threading.current_thread().ident
        if i in self.buff_dict:
            self.buff_dict[i].truncate(*args, **kwargs)

    def free(self):
        i = threading.current_thread().ident
        if i in self.buff_dict:
            buff = self.buff_dict.pop(threading.current_thread().ident)
            buff.seek(0)
            buff.truncate()
            self.buff_queue.append(buff)


_io_buffer = _PooledIOBuffer()
out = _io_buffer

PASS = 0
SKIPPED = 1
ERROR = 2
FAIL = 3
EXPECTED_FAIL = 4
UNEXPECTED_PASS = 5


class BufferedTestResult(unittest.TextTestResult):
    def __init__(self, stream, descriptions, verbosity):
        super(BufferedTestResult, self).__init__(stream, descriptions, verbosity)
        self.buffer = False
        # key: module_name.class_name, value: a list of results.
        # One result is a tuple like (test method name, method doc, duration, status, output, traceback)
        self.results = dict()
        self.start_time = 0.0
        self.sum_duration = 0.0
        self.successes = 0
        self.name = 'test'
        self.description = ''
        self.children = []

    def __len__(self):
        return len(self.children)

    def __iter__(self):
        return iter(self.children)

    def _add_result(self, test, duration, status, output, trace_back):
        mod_name = get_module_name(test)
        cls_name = get_long_class_name(test)
        if mod_name not in self.results:
            self.results[mod_name] = dict()
        if cls_name not in self.results[mod_name]:
            self.results[mod_name][cls_name] = []
        test_name, test_doc = self.__class__._get_test_info(test)
        output = output or 'No Log\n'
        trace_back = trace_back or 'No Exception\n'
        self.results[mod_name][cls_name].append((test_name, test_doc, duration, status, output, trace_back))

    @staticmethod
    def _get_test_info(test):
        test_name = get_long_method_name(test)
        test_doc = getattr(test, '_testMethodDoc', None)
        return test_name, test_doc or 'No Method Doc\n'

    def _exc_info_to_string(self, error, test):
        """Almost the same as its base class implementation, except eliminating the mirror output"""
        exctype, value, tb = error
        # Skip test runner traceback levels
        while tb and self._is_relevant_tb_level(tb):
            tb = tb.tb_next
        if exctype is test.failureException:
            # Skip assert*() traceback levels
            length = self._count_relevant_tb_levels(tb)
            msg_lines = traceback.format_exception(exctype, value, tb, length)
        else:
            msg_lines = traceback.format_exception(exctype, value, tb)
        return ''.join(msg_lines)

    def startTest(self, test):
        super(BufferedTestResult, self).startTest(test)
        self.start_time = time.time()

    def stopTest(self, test):
        self._mirrorOutput = False
        _io_buffer.free()

    def addSuccess(self, test):
        duration = time.time() - self.start_time
        super(BufferedTestResult, self).addSuccess(test)
        self.successes += 1
        self._add_result(test, duration, PASS, _io_buffer.getvalue(), '')

    def addError(self, test, error):
        duration = time.time() - self.start_time
        super(BufferedTestResult, self).addError(test, error)
        test_obj, exception_str = self.errors[-1]
        self._add_result(test, duration, ERROR, _io_buffer.getvalue(), exception_str)

    def addFailure(self, test, error):
        duration = time.time() - self.start_time
        super(BufferedTestResult, self).addFailure(test, error)
        test_obj, exception_str = self.failures[-1]
        self._add_result(test, duration, FAIL, _io_buffer.getvalue(), exception_str)

    def addSkip(self, test, reason):
        duration = time.time() - self.start_time
        super(BufferedTestResult, self).addSkip(test, reason)
        test_obj, reason = self.skipped[-1]
        self._add_result(test, duration, SKIPPED, _io_buffer.getvalue(), 'Skipped: {0!r}'.format(reason))

    def addExpectedFailure(self, test, error):
        duration = time.time() - self.start_time
        super(BufferedTestResult, self).addExpectedFailure(test, error)
        test_obj, exception_str = self.expectedFailures[-1]
        self._add_result(test, duration, EXPECTED_FAIL, _io_buffer.getvalue(), exception_str)

    def addUnexpectedSuccess(self, test):
        duration = time.time() - self.start_time
        super(BufferedTestResult, self).addUnexpectedSuccess(test)
        self._add_result(test, duration, UNEXPECTED_PASS, _io_buffer.getvalue(), '')

    def wasSuccessful(self):
        return len(self.failures) == len(self.errors) == len(self.unexpectedSuccesses) == 0


def combine_results(result, results):
    for r in results:
        result.failures.extend(r.failures)
        result.errors.extend(r.errors)
        result.testsRun += r.testsRun
        result.skipped.extend(r.skipped)
        result.expectedFailures.extend(r.expectedFailures)
        result.unexpectedSuccesses.extend(r.unexpectedSuccesses)
        result.successes += r.successes
        for mod_name, mod in r.results.items():
            if mod_name not in result.results:
                result.results[mod_name] = dict()
            for cls_name, tups in mod.items():
                if cls_name not in result.results[mod_name]:
                    result.results[mod_name][cls_name] = []
                result.results[mod_name][cls_name].extend(tups)


# Copyright 2015 Twitter, Inc and other contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import unishark
import sys
import logging
import concurrent.futures
import time


log = logging.getLogger(__name__)


class TestProgram(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def run(self):
        pass


class DefaultTestProgram(TestProgram):
    def __init__(self, test_dict_conf, verbosity=1, descriptions=False):
        self.test_dict_conf = test_dict_conf
        self.name_pattern = None
        if 'name_pattern' in self.test_dict_conf['test']:
            self.name_pattern = self.test_dict_conf['test']['name_pattern']
        self.verbosity = verbosity
        self.descriptions = descriptions
        self.reporters = self._make_reporters()

    def run(self):
        test = self.test_dict_conf['test']
        if 'max_workers' in test:  # Deprecation message
            raise KeyError('Please set "max_workers" in the "concurrency" sub-dict instead.')
        concurrency = test['concurrency'] if 'concurrency' in test else {
            'max_workers': 1,
            'timeout': None
        }
        int(concurrency['max_workers'])  # if concurrency key exists, max_workers is required and must be int
        if 'timeout' not in concurrency:
            concurrency['timeout'] = None
        if concurrency['max_workers'] <= 1:
            return self._run_suites_sequentially()
        else:
            return self._run_suites_concurrently(concurrency['max_workers'], concurrency['timeout'])

    @staticmethod
    def _get_class_from_name(long_cls_name):
        parts = long_cls_name.split('.')
        mod = __import__('.'.join(parts[:-1]))
        cls = getattr(mod, parts[-1], None)
        if cls is None:
            raise AttributeError('Class not found.')
        if not issubclass(cls, unishark.Reporter):
            raise TypeError('Class is not a subclass of Reporter.')
        return cls

    def _make_reporters(self):
        created_reporters = []
        test = self.test_dict_conf['test']
        reporters = self.test_dict_conf['reporters'] if 'reporters' in self.test_dict_conf else dict()
        if 'reporters' in test:
            for name in test['reporters']:
                reporter = reporters[name]
                reporter_class = self.__class__._get_class_from_name(reporter['class'])
                kwargs = reporter['kwargs'] if 'kwargs' in reporter else None
                if kwargs and type(kwargs) is dict:
                    created_reporters.append(reporter_class(**kwargs))
                else:
                    created_reporters.append(reporter_class())
        return created_reporters

    def _run_suites_sequentially(self):
        exit_code = 0
        suites = unishark.DefaultTestLoader(name_pattern=self.name_pattern).load_tests_from_dict(self.test_dict_conf)
        runner = unishark.BufferedTestRunner(reporters=self.reporters,
                                             verbosity=self.verbosity,
                                             descriptions=self.descriptions)
        for suite_name, suite_content in suites.items():
            package_name = suite_content['package']
            suite = suite_content['suite']
            concurrency = suite_content['concurrency']
            result = runner.run(suite, name=suite_name, description='Package: ' + package_name,
                                max_workers=concurrency['max_workers'], concurrency_level=concurrency['level'],
                                timeout=concurrency['timeout'])
            exit_code += 0 if result.wasSuccessful() else 1
        for reporter in self.reporters:
            reporter.collect()
        return exit_code

    def _run_suites_concurrently(self, max_workers_on_suites, timeout):
        exit_code = 0
        suites = unishark.DefaultTestLoader(name_pattern=self.name_pattern).load_tests_from_dict(self.test_dict_conf)
        start_time = time.time()
        with concurrent.futures.ThreadPoolExecutor(max_workers_on_suites) as executor:
            futures = []
            for suite_name, suite_content in suites.items():
                package_name = suite_content['package']
                suite = suite_content['suite']
                concurrency = suite_content['concurrency']
                runner = unishark.BufferedTestRunner(reporters=self.reporters,
                                                     verbosity=self.verbosity,
                                                     descriptions=self.descriptions)
                future = executor.submit(runner.run, suite,
                                         name=suite_name,
                                         description='Package: ' + package_name,
                                         max_workers=concurrency['max_workers'],
                                         concurrency_level=concurrency['level'],
                                         timeout=concurrency['timeout'])
                futures.append(future)
            for future in concurrent.futures.as_completed(futures, timeout=timeout):
                result = future.result()
                exit_code += 0 if result.wasSuccessful() else 1
        actual_duration = time.time() - start_time
        log.info('Actual total time taken: %.3fs' % actual_duration)
        for reporter in self.reporters:
            reporter.set_actual_duration(actual_duration)
            reporter.collect()
        return exit_code


def main(test_program):
    if not isinstance(test_program, TestProgram):
        raise TypeError
    exit_code = test_program.run()
    log.info('Exit Code: %d' % exit_code)
    sys.exit(exit_code)


# Copyright 2015 Twitter, Inc and other contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['DefaultTestLoader',
           'out', 'BufferedTestResult', 'BufferedTestRunner',
           'HtmlReporter', 'XUnitReporter',
           'main', 'DefaultTestProgram',
           'data_driven', 'multi_threading_data_driven',
           'ContextManager', 'contexts']

from unishark.result import (out, BufferedTestResult)
from unishark.reporter import (Reporter, HtmlReporter, XUnitReporter)
from unishark.runner import BufferedTestRunner
from unishark.loader import DefaultTestLoader
from unishark.decorator import data_driven, multi_threading_data_driven
from unishark.util import ContextManager, contexts
from unishark.main import (TestProgram, DefaultTestProgram, main)


PACKAGE = __name__
VERSION = '0.3.1'



# Copyright 2015 Twitter, Inc and other contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unishark.suite import TestSuite, convert
from unishark.result import BufferedTestResult, out
from unishark.reporter import Reporter
from unittest.signals import registerResult
import time
import warnings
import logging


log = logging.getLogger(__name__)

_concurrency_level_to_int = {
    'none': TestSuite.ROOT_LEVEL,
    'module': TestSuite.MODULE_LEVEL,
    'class': TestSuite.CLASS_LEVEL,
    'method': TestSuite.METHOD_LEVEL
}


class BufferedTestRunner(unittest.TextTestRunner):
    def __init__(self, reporters=None, verbosity=1, descriptions=False):
        super(BufferedTestRunner, self).__init__(buffer=False,
                                                 verbosity=verbosity,
                                                 descriptions=descriptions,
                                                 resultclass=BufferedTestResult)
        if reporters:
            self.reporters = reporters
        else:
            self.reporters = []
        for reporter in self.reporters:
            if not isinstance(reporter, Reporter):
                raise TypeError

    def make_result(self):
        return self.resultclass(self.stream, self.descriptions, self.verbosity)

    def make_results_tree(self, test, result):
        if self.__class__._is_suite(test):
            result.children = [self.make_result() for _ in test]
            for t, r in zip(test, result.children):
                self.make_results_tree(t, r)

    def _before_run(self):
        # Keep the same as lines 145-162 in unittest.TextTextRunner.run
        result = self.make_result()
        registerResult(result)
        result.failfast = self.failfast
        result.buffer = self.buffer
        with warnings.catch_warnings():
            warn = getattr(self, 'warnings', None)
            if warn:
                warnings.simplefilter(warn)
                if warn in ['default', 'always']:
                    warnings.filterwarnings('module',
                                            category=DeprecationWarning,
                                            message='Please use assert\w+ instead.')
        return result

    def _after_run(self, result):
        # Almost the same as lines 175-213 in unittest.TextTextRunner.run,
        # with small fix of counting unexpectedSuccesses into a FAILED run.
        result.printErrors()
        if hasattr(result, 'separator2'):
            self.stream.writeln(result.separator2)
        run = result.testsRun
        self.stream.writeln("Ran %d test%s in %.3fs" %
                            (run, run != 1 and "s" or "", result.sum_duration))
        self.stream.writeln()

        expected = unexpected = skipped = 0
        try:
            results = map(len, (result.expectedFailures,
                                result.unexpectedSuccesses,
                                result.skipped))
        except AttributeError:
            pass
        else:
            expected, unexpected, skipped = results

        infos = []
        if not result.wasSuccessful():
            self.stream.write("FAILED")
            failed, errored = len(result.failures), len(result.errors)
            if failed:
                infos.append("failures=%d" % failed)
            if errored:
                infos.append("errors=%d" % errored)
            if unexpected:
                infos.append("unexpected successes=%d" % unexpected)
        else:
            self.stream.write("OK")
        if skipped:
            infos.append("skipped=%d" % skipped)
        if expected:
            infos.append("expected failures=%d" % expected)
        if infos:
            self.stream.writeln(" (%s)" % (", ".join(infos),))
        else:
            self.stream.write("\n")

    @staticmethod
    def _is_suite(test):
        try:
            iter(test)
        except TypeError:
            return False
        return True

    def run(self, test, name='test', description='', max_workers=1, concurrency_level='class', timeout=None):
        result = self._before_run()
        result.name = name
        result.description = description
        start_time = time.time()
        start_test_run = getattr(result, 'startTestRun', None)
        if start_test_run is not None:
            start_test_run()
        try:
            if not self.__class__._is_suite(test):
                test(result)
            else:
                test = convert(test)
                self.make_results_tree(test, result)
                test.run(result, concurrency_level=_concurrency_level_to_int[concurrency_level],
                         max_workers=max_workers, timeout=timeout)
        finally:
            stop_test_run = getattr(result, 'stopTestRun', None)
            if stop_test_run is not None:
                stop_test_run()
        result.sum_duration = time.time() - start_time
        self._after_run(result)

        log.debug('%d free IO buffer(s) in the pool, %d in use.' %
                  (len(out.buff_queue), len(out.buff_dict)))
        if out.buff_dict:
            log.debug('IO buffers in use: %r' % out.buff_dict)

        for reporter in self.reporters:
            reporter.report(result)
        return result


# Copyright 2015 Twitter, Inc and other contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class MultipleErrors(Exception):
    def __init__(self, msgs):
        self.msgs = msgs

    def __len__(self):
        return len(self.msgs)

    def __str__(self):
        return 'Total %d error(s) as follows: \n\n%s' % (len(self.msgs), '\n\n'.join(self.msgs))

    def __repr__(self):
        return repr(self.msgs)


# Copyright 2015 Twitter, Inc and other contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import inspect
from os import sep
from traceback import format_exception


class ContextManager(object):
    _shared_state = dict()

    def __init__(self):
        self.__dict__ = self._shared_state
        self._context_dict = dict()

    def get(self, name):
        return self._context_dict[name]

    def set(self, name, context):
        self._context_dict[name] = context


contexts = ContextManager()


def get_module_name(obj):
    return inspect.getsourcefile(type(obj)).split(sep)[-1].rstrip('.py')


def get_long_class_name(obj):
    return '.'.join((get_module_name(obj), get_class_name(obj)))


def get_long_method_name(obj):
    return '.'.join((get_long_class_name(obj), get_method_name(obj)))


def get_method_name(obj):
    return obj.id().split('.')[-1]


def get_class_name(obj):
    return type(obj).__name__


def _is_relevant_tb_level(tb):
    return '__unittest' in tb.tb_frame.f_globals


def _count_relevant_tb_levels(tb):
    length = 0
    while tb and not _is_relevant_tb_level(tb):
        length += 1
        tb = tb.tb_next
    return length


def exc_info_to_string(exc_info):
    """Also see runner.BufferedTestResult"""
    exctype, value, tb = exc_info
    # Skip test runner traceback levels
    while tb and _is_relevant_tb_level(tb):
        tb = tb.tb_next
    if exctype is AssertionError:
        # Skip assert*() traceback levels
        length = _count_relevant_tb_levels(tb)
        msg_lines = format_exception(exctype, value, tb, length)
    else:
        msg_lines = format_exception(exctype, value, tb)
    return ''.join(msg_lines).rstrip()


# Copyright 2015 Twitter, Inc and other contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import pyclbr
import logging
import types
import re

log = logging.getLogger(__name__)


class DefaultTestLoader:
    def __init__(self, name_pattern=None):
        self._name_tree = None
        self._case_class = unittest.TestCase
        self._suite_class = unittest.TestSuite
        self._name_trees_by_pkg = dict()
        self.name_pattern = name_pattern or r'^test\w*'
        self.one_dot_name_pattern = r'\w+\.test\w*'
        self.two_dots_name_pattern = r'(\w+\.){2}test\w*'

    def load_tests_from_dict(self, dict_conf):
        suites_dict = dict()
        for suite_name, content in self._parse_tests_from_dict(dict_conf).items():
            package = content['package']
            test_case_names = content['test_case_names']
            suite = self.load_tests_from_full_names(test_case_names)
            if suite.countTestCases() <= 0:
                log.error('Test suite %r is empty.' % suite_name)
            suites_dict[suite_name] = {
                'package': package,
                'suite': suite,
                'concurrency': content['concurrency']
            }
            log.info('Created test suite %r successfully from package %r.' % (suite_name, package))
        return suites_dict

    def load_tests_from_full_names(self, full_names):
        """
        Returns a unittest.TestSuite instance containing the tests
        whose full name is given and short method name filtered by name_pattern.
        A full name is a dotted name like (package.)module.class.method
        """
        full_names = self._filter_tests_by_name_pattern(full_names)
        return self._make_suite_from_full_names(full_names)

    def load_tests_from_package(self, pkg_name, regex=None):
        """
        Returns a unittest.TestSuite instance containing the tests
        whose dotted long name module.class.method matches the given regular expression
        and short method name matches name_pattern.
        A dotted package name must be provided. regex is default to '(\w+\.){2}test\w*'
        """
        self._build_pkg_name_tree(pkg_name)
        full_names = self._get_full_method_names_from_tree(pkg_name)
        full_names = self._filter_tests_by_two_dots_name_pattern(full_names, regex=regex)
        full_names = self._filter_tests_by_name_pattern(full_names)
        return self._make_suite_from_full_names(full_names)

    def load_tests_from_modules(self, mod_names, regex=None):
        """
        Returns a unittest.TestSuite instance containing the tests
        whose dotted long name class.method matches the given regular expression
        and short method name matches name_pattern.
        A list of dotted module names must be provided. regex is default to '\w+\.test\w*'
        """
        self._build_name_tree(None, set(mod_names))
        full_names = self._get_full_method_names_from_tree(None)
        full_names = self._filter_tests_by_one_dot_name_pattern(full_names, regex=regex)
        full_names = self._filter_tests_by_name_pattern(full_names)
        return self._make_suite_from_full_names(full_names)

    def _filter_tests_by_name_pattern(self, full_names):
        return list(filter(lambda n: re.match(self.name_pattern, n.split('.')[-1]), full_names))

    def _filter_tests_by_one_dot_name_pattern(self, full_names, regex=None):
        regex = regex or self.one_dot_name_pattern
        return list(filter(lambda n: re.match(regex, '.'.join(n.split('.')[-2:])), full_names))

    def _filter_tests_by_two_dots_name_pattern(self, full_names, regex=None):
        regex = regex or self.two_dots_name_pattern
        return list(filter(lambda n: re.match(regex, '.'.join(n.split('.')[-3:])), full_names))

    def _make_suite_from_full_names(self, full_names):
        cases = list(filter(lambda c: c is not None, [self._make_case_from_full_name(name) for name in full_names]))
        suite = self._suite_class(cases)
        log.info('Loaded %d test(s).' % suite.countTestCases())
        log.debug('Loaded test(s): %r' % suite)
        return suite

    def _make_case_from_full_name(self, full_name):
        name_parts = full_name.split('.')
        mod = None
        importable_parts = name_parts[:]
        while importable_parts:
            try:
                mod = __import__('.'.join(importable_parts))
                break
            except ImportError:
                del importable_parts[-1]
                if not importable_parts:
                    raise
        obj = mod
        parent = None
        for part in name_parts[1:]:
            parent, obj = obj, getattr(obj, part)
        if not isinstance(obj, types.FunctionType) and not isinstance(obj, types.MethodType):
            raise TypeError('%r is neither %r nor %r' % (obj, types.FunctionType, types.MethodType))
        # filter out class if it is not unittest.TestCase subclass
        elif isinstance(parent, type) and issubclass(parent, self._case_class):
            name = name_parts[-1]
            test_obj = parent(name)
            # filter out static methods
            if not isinstance(getattr(test_obj, name), types.FunctionType):
                return test_obj
        else:
            return None

    def _parse_tests_from_dict(self, dict_conf):
        res_suites = dict()
        test = dict_conf['test']
        suites = dict_conf['suites']
        for suite_name in test['suites']:
            suite = suites[suite_name]
            pkg_name = None
            if 'package' in suite and suite['package']:
                pkg_name = suite['package']
            groups_dict = suite['groups']
            test_cases_names = []
            for group_key, group in groups_dict.items():
                group = groups_dict[group_key]
                if 'disable' in group and group['disable']:
                    continue
                gran = group['granularity']
                if gran == 'package':
                    full_mth_names = self._get_full_method_names_from_package(pkg_name, group)
                    test_cases_names.extend(full_mth_names)
                elif gran == 'module':
                    full_mth_names = self._get_full_method_names_from_modules(pkg_name, group)
                    test_cases_names.extend(full_mth_names)
                elif gran == 'class':
                    full_mth_names = self._get_full_method_names_from_classes(pkg_name, group)
                    test_cases_names.extend(full_mth_names)
                elif gran == 'method':
                    full_mth_names = self._get_full_method_names_from_methods(pkg_name, group)
                    test_cases_names.extend(full_mth_names)
                else:
                    raise ValueError('Granularity must be one of %r.' % ['package', 'module', 'class', 'method'])
            if 'max_workers' in suite or 'concurrency_level' in suite:  # Deprecation message
                raise KeyError('Please set "max_workers" and "level" in the "concurrency" sub-dict instead.')
            concurrency = self.__class__._parse_concurrency_conf(suite)
            res_suites[suite_name] = {
                'package': pkg_name or 'None',
                'test_case_names': set(test_cases_names),
                'concurrency': concurrency
            }
        log.debug('Parsed test config: %r' % res_suites)
        log.info('Parsed test config successfully.')
        return res_suites

    @staticmethod
    def _parse_concurrency_conf(suite_conf):
        concurrency = suite_conf['concurrency'] if 'concurrency' in suite_conf else {
            'max_workers': 1,
            'level': 'class',
            'timeout': None
        }
        int(concurrency['max_workers'])  # if concurrency key exists, max_workers is required and must be int
        if 'level' not in concurrency:
            concurrency['level'] = 'class'
        concur_levels = ['module', 'class', 'method']
        if concurrency['level'] not in concur_levels:
            raise ValueError('Concurrency level (%r) is not one of %r.' % (concurrency['level'], concur_levels))
        if 'timeout' not in concurrency:
            concurrency['timeout'] = None
        return concurrency

    def _build_pkg_name_tree(self, pkg_name):
        import pkgutil
        import copy
        if not pkg_name:
            raise ValueError('A dotted package name must be provided.')
        if pkg_name in self._name_trees_by_pkg:
            self._name_tree = copy.deepcopy(self._name_trees_by_pkg[pkg_name])
        else:
            pkg = __import__(pkg_name)
            name_parts = pkg_name.split('.')
            if len(name_parts) > 1:
                for part in name_parts[1:]:
                    pkg = getattr(pkg, part)
            members = pkgutil.iter_modules(pkg.__path__)
            mod_names = []
            for _, mod_name, is_pkg in members:
                if not is_pkg:
                    mod_names.append(mod_name)
            self._build_name_tree(pkg_name, mod_names)
            self._name_trees_by_pkg[pkg_name] = copy.deepcopy(self._name_tree)

    def _get_full_method_names_from_package(self, pkg_name, group):
        self._build_pkg_name_tree(pkg_name)
        if 'except_modules' in group:
            self._del_except_modules_in_name_tree(group['except_modules'])
        if 'except_classes' in group:
            self._del_except_classes_in_name_tree(group['except_classes'])
        if 'except_methods' in group:
            self._del_except_methods_in_name_tree(group['except_methods'])
        full_names = self._get_full_method_names_from_tree(pkg_name)
        pattern = group.get('pattern', None)
        return self._filter_tests_by_two_dots_name_pattern(full_names, regex=pattern)

    def _get_full_method_names_from_modules(self, pkg_name, group):
        mod_names = group['modules']
        mod_names = set(mod_names)
        self._build_name_tree(pkg_name, mod_names)
        if 'except_classes' in group:
            self._del_except_classes_in_name_tree(group['except_classes'])
        if 'except_methods' in group:
            self._del_except_methods_in_name_tree(group['except_methods'])
        return self._get_full_method_names_from_tree(pkg_name)

    def _get_full_method_names_from_classes(self, pkg_name, group):
        long_cls_names = group['classes']
        long_cls_names = set(long_cls_names)
        mod_names = set()
        for long_cls_name in long_cls_names:
            mod_name, _ = self.__class__._get_cls_name_parts(long_cls_name)
            mod_names.add(mod_name)
        self._build_name_tree(pkg_name, mod_names, filter_cls_names=long_cls_names)
        if 'except_methods' in group:
            self._del_except_methods_in_name_tree(group['except_methods'])
        return self._get_full_method_names_from_tree(pkg_name)

    def _get_full_method_names_from_methods(self, pkg_name, group):
        long_mth_names = group['methods']
        long_mth_names = set(long_mth_names)
        for long_mth_name in long_mth_names:
            self.__class__._get_mth_name_parts(long_mth_name)
        full_mth_names = list(map(lambda n: '.'.join((pkg_name, n)), long_mth_names)) \
            if pkg_name else long_mth_names
        return full_mth_names

    @staticmethod
    def _get_cls_name_parts(long_name):
        parts = long_name.split('.')
        if len(parts) != 2:
            raise ValueError('%r does not comply with: "module.class".' % long_name)
        return parts[0], parts[1]

    @staticmethod
    def _get_mth_name_parts(long_name):
        parts = long_name.split('.')
        if len(parts) != 3:
            raise ValueError('%r does not comply with: "module.class.method".' % long_name)
        return parts[0], parts[1], parts[2]

    def _del_except_modules_in_name_tree(self, except_mod_names):
        for mod_name in set(except_mod_names):
            self._del_mod_in_name_tree(mod_name)

    def _del_except_classes_in_name_tree(self, except_cls_names):
        for except_cls_name in set(except_cls_names):
            mod_name, cls_name = self.__class__._get_cls_name_parts(except_cls_name)
            self._del_cls_in_name_tree(mod_name, cls_name)

    def _del_except_methods_in_name_tree(self, except_mth_names):
        for except_mth_name in set(except_mth_names):
            mod_name, cls_name, mth_name = self.__class__._get_mth_name_parts(except_mth_name)
            self._del_mth_in_name_tree(mod_name, cls_name, mth_name)

    @staticmethod
    def _inspect_module(pkg_name, mod_name):
        full_mod_name = mod_name if not pkg_name else '.'.join((pkg_name, mod_name))
        return pyclbr.readmodule_ex(full_mod_name)  # return module_content

    @staticmethod
    def _get_classes_of_module(module_content):
        classes = dict()  # a dict of 'class_name': pyclbr.Class obj
        for name, obj in module_content.items():
            if isinstance(obj, pyclbr.Class):
                classes[name] = obj
        return classes

    @staticmethod
    def _get_method_names_by_class(cls):
        # cls.methods is a dict of 'method_name': method_line_no
        return sorted(cls.methods.keys(), key=lambda k: cls.methods[k])

    # A name tree is like:
    # tree = {
    #     'mod1': {
    #         'cls1': {
    #             'mth1', 'mth2', ...
    #         },
    #         'cls2': {
    #             ...
    #         },
    #         ...
    #     },
    #     'mod2': {
    #         ...
    #     },
    #     ...
    # }
    def _build_name_tree(self, pkg_name, mod_names, filter_cls_names=None):
        tree = self._name_tree = dict()
        for mod_name in mod_names:
            full_mod_name = '.'.join((pkg_name, mod_name)) if pkg_name else mod_name
            __import__(full_mod_name)
        for mod_name in mod_names:
            mod_content = self.__class__._inspect_module(pkg_name, mod_name)
            classes = self.__class__._get_classes_of_module(mod_content)
            if not classes:
                continue
            for cls_name, cls in classes.items():
                long_cls_name = '.'.join((mod_name, cls_name))
                if filter_cls_names and long_cls_name not in filter_cls_names:
                    continue
                mth_names = self.__class__._get_method_names_by_class(cls)
                if not mth_names:
                    continue
                if mod_name not in tree:
                    tree[mod_name] = dict()
                if cls_name not in tree[mod_name]:
                    tree[mod_name][cls_name] = set()
                for mth_name in mth_names:
                    tree[mod_name][cls_name].add(mth_name)

    def _del_mod_in_name_tree(self, mod_name):
        if mod_name not in self._name_tree:
            raise ValueError('Cannot exclude %r because it is not included.' % mod_name)
        del self._name_tree[mod_name]

    def _del_cls_in_name_tree(self, mod_name, cls_name):
        long_cls_name = '.'.join((mod_name, cls_name))
        if mod_name not in self._name_tree or cls_name not in self._name_tree[mod_name]:
            raise ValueError('Cannot exclude %r because it is not included.' % long_cls_name)
        del self._name_tree[mod_name][cls_name]
        if not self._name_tree[mod_name]:
            del self._name_tree[mod_name]

    def _del_mth_in_name_tree(self, mod_name, cls_name, mth_name):
        long_mth_name = '.'.join((mod_name, cls_name, mth_name))
        if mod_name not in self._name_tree \
                or cls_name not in self._name_tree[mod_name] \
                or mth_name not in self._name_tree[mod_name][cls_name]:
            raise ValueError('Cannot exclude %r because it is not included.' % long_mth_name)
        self._name_tree[mod_name][cls_name].remove(mth_name)
        if not self._name_tree[mod_name][cls_name]:
            del self._name_tree[mod_name][cls_name]
        if not self._name_tree[mod_name]:
            del self._name_tree[mod_name]

    def _get_full_method_names_from_tree(self, pkg_name):
        full_method_names = []
        self._get_dotted_names_dfs(self._name_tree, pkg_name, full_method_names)
        return full_method_names

    def _get_dotted_names_dfs(self, tree, prefix, dotted_names):
        for name in tree:
            dotted_name = '.'.join((prefix, name)) if prefix else name
            if type(tree) is set:  # leaf
                dotted_names.append(dotted_name)
            else:  # tree is dict
                self._get_dotted_names_dfs(tree[name], dotted_name, dotted_names)


# Copyright 2015 Twitter, Inc and other contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jinja2
import re
import os
import codecs
import datetime
import abc
import unishark
from unishark.result import (PASS, SKIPPED, ERROR, FAIL, EXPECTED_FAIL, UNEXPECTED_PASS)
import threading

_status_to_str = {
    PASS: 'Passed',
    SKIPPED: 'Skipped',
    ERROR: 'Error',
    FAIL: 'Failed',
    EXPECTED_FAIL: 'Passed',
    UNEXPECTED_PASS: 'Failed'
}


class Reporter(object):
    """Base class of all reporter classes"""
    __metaclass__ = abc.ABCMeta

    lock = threading.RLock()

    def __init__(self):
        self._actual_duration = None

    def set_actual_duration(self, duration):
        self._actual_duration = duration

    @abc.abstractmethod
    def report(self, result):
        pass

    @abc.abstractmethod
    def collect(self):
        pass


class Summary(object):
    """Base class of representation classes"""
    def __init__(self, name):
        self.name = name
        self.description = ''
        self.duration = 0.0
        self.run = 0
        self.passed = 0
        self.skipped = 0
        self.error = 0
        self.fail = 0
        self.rate = ''
        self.category = ''

    def round_duration(self):
        self.duration = round(self.duration, 3)

    def calc_rate(self):
        if self.run:
            self.rate = str(round((self.passed + self.skipped) * 100.0 / self.run, 2)) + '%'
        else:
            self.rate = '0%'

    def calc_category(self):
        self.category = (self.fail or self.error) and 'danger' or self.skipped and 'warning' or 'success'


class TestsSummary(Summary):
    def __init__(self, name):
        super(TestsSummary, self).__init__(name)
        self.suite_sum_list = []

    def build(self, actual_duration=None):
        if self.suite_sum_list:
            if actual_duration is not None:
                self.duration = actual_duration
            else:
                self.duration = sum(map(lambda s: s.duration, self.suite_sum_list))
            self.run = sum(map(lambda s: s.run, self.suite_sum_list))
            self.passed = sum(map(lambda s: s.passed, self.suite_sum_list))
            self.skipped = sum(map(lambda s: s.skipped, self.suite_sum_list))
            self.error = sum(map(lambda s: s.error, self.suite_sum_list))
            self.fail = sum(map(lambda s: s.fail, self.suite_sum_list))
            self.round_duration()
            self.calc_category()
            self.calc_rate()


class SuiteSummary(Summary):
    def __init__(self, name):
        super(SuiteSummary, self).__init__(name)
        self.mod_sum_list = []

    def build(self, result):
        # Suite summary
        self.description = result.description
        self.duration = result.sum_duration
        self.run = result.testsRun
        self.passed = result.successes + len(result.expectedFailures)
        self.skipped = len(result.skipped)
        self.error = len(result.errors)
        self.fail = len(result.failures) + len(result.unexpectedSuccesses)
        self.round_duration()
        self.calc_category()
        self.calc_rate()
        # Module summary
        mod_id = cls_id = mth_id = 0
        for mod_name in result.results:
            mod_id += 1
            mod_results = result.results[mod_name]
            mod_sum = ModuleSummary(mod_id, mod_name)
            # Class summary
            for cls_name in mod_results:
                cls_id += 1
                result_list = mod_results[cls_name]
                cls_sum = ClassSummary(cls_id, cls_name)
                for r in result_list:
                    # Method summary
                    mth_id += 1
                    mth_name, mth_doc, mth_duration, mth_status, mth_output, mth_traceback = r
                    mth_sum = MethodSummary(mth_id, mth_name, mth_duration, mth_status,
                                            mth_doc, mth_output, mth_traceback)
                    # Fill stats in class summary
                    cls_sum.mth_sum_list.append(mth_sum)
                    cls_sum.run += 1
                    cls_sum.duration += mth_duration
                    if mth_status == EXPECTED_FAIL or mth_status == PASS:
                        cls_sum.passed += 1
                    elif mth_status == UNEXPECTED_PASS or mth_status == FAIL:
                        cls_sum.fail += 1
                    elif mth_status == SKIPPED:
                        cls_sum.skipped += 1
                    else:
                        cls_sum.error += 1
                cls_sum.round_duration()
                cls_sum.calc_category()
                cls_sum.calc_rate()
                # Skip filling stats in module summary
                mod_sum.cls_sum_list.append(cls_sum)
            self.mod_sum_list.append(mod_sum)


class ModuleSummary(Summary):
    def __init__(self, mid, name):
        super(ModuleSummary, self).__init__(name)
        self.mid = 'mod-' + str(mid)
        self.cls_sum_list = []


class ClassSummary(Summary):
    def __init__(self, cid, name):
        super(ClassSummary, self).__init__(name)
        self.cid = 'cls-' + str(cid)
        self.mth_sum_list = []


class MethodSummary(Summary):
    def __init__(self, fid, name, duration, status, doc, output, trace_back):
        super(MethodSummary, self).__init__(name)
        self.fid = 'mth-' + str(fid)
        self.duration = duration
        self.round_duration()
        self.status = _status_to_str[status]
        self.category = (status == PASS or status == EXPECTED_FAIL) and 'pass' \
            or (status == FAIL or status == UNEXPECTED_PASS) and 'fail' \
            or status == SKIPPED and 'skipped' \
            or 'error'
        self.description = doc
        self.output = output
        self.trace_back = trace_back


class HtmlReporter(Reporter):
    def __init__(self, dest='results',
                 overview_title='Reports',
                 overview_description='',
                 templates_path=None,
                 report_template=None,
                 overview_template=None,
                 index_template=None):
        super(HtmlReporter, self).__init__()
        self.dest = dest
        self._tests_sum = TestsSummary(overview_title)
        self._tests_sum.description = overview_description
        if templates_path and report_template and overview_template and index_template:
            loader = jinja2.FileSystemLoader(templates_path, encoding='utf-8')
            self.report_template = report_template
            self.overview_template = overview_template
            self.index_template = index_template
        else:
            loader = jinja2.PackageLoader(unishark.PACKAGE, package_path='templates', encoding='utf-8')
            self.report_template = 'report.html'
            self.overview_template = 'overview.html'
            self.index_template = 'index.html'
        self.jinja_env = jinja2.Environment(loader=loader, autoescape=False)

        # Add converting newline to <br> filter
        @jinja2.evalcontextfilter
        def nl2br(eval_ctx, text):
            paragraph_re = re.compile(r'(?:\r\n|\r|\n){2,}')
            result = u'\n\n'.join(paragraph_re.split(jinja2.escape(text)))
            result = result.replace('\n', '<br/>\n')
            if eval_ctx.autoescape:
                result = jinja2.Markup(result)
            return result
        self.jinja_env.filters['nl2br'] = nl2br

        # Add <pre> tag wrapper filter
        @jinja2.evalcontextfilter
        def pre(eval_ctx, value):
            _paragraph_re = re.compile(r'(?:\r\n|\r|\n){2,}')
            text = u'\n\n'.join(u'<pre>%s</pre>' % p
                                for p in _paragraph_re.split(jinja2.escape(value)))
            if eval_ctx.autoescape:
                text = jinja2.Markup(text)
            return text
        self.jinja_env.filters['pre'] = pre

    def _make_output_dir(self):
        with Reporter.lock:
            if not os.path.exists(self.dest):
                os.makedirs(self.dest)

    def report(self, result):
        self._make_output_dir()
        # Heading
        date_time = str(datetime.datetime.now())
        app = unishark.PACKAGE
        version = unishark.VERSION
        # Summary
        suite_sum = SuiteSummary(result.name)
        suite_sum.build(result)
        # Generate report from template
        template = self.jinja_env.get_template(self.report_template)
        html = template.render(app=app,
                               version=version,
                               date_time=date_time,
                               suite_sum=suite_sum)
        filename = os.path.join(self.dest, suite_sum.name + '_result.html')
        with codecs.open(filename, encoding='utf-8', mode='w+') as f:
            f.write(html)

        self._tests_sum.suite_sum_list.append(suite_sum)

    def collect(self):
        self._make_output_dir()
        self._generate_overview()
        self._generate_index()

    def _generate_overview(self):
        # Overview heading
        date_time = str(datetime.datetime.now())
        app = unishark.PACKAGE
        version = unishark.VERSION
        # Summary
        self._tests_sum.build(actual_duration=self._actual_duration)
        # Generate overview from template
        template = self.jinja_env.get_template(self.overview_template)
        html = template.render(app=app,
                               version=version,
                               date_time=date_time,
                               tests_sum=self._tests_sum)
        with codecs.open(os.path.join(self.dest, 'overview.html'), encoding='utf-8', mode='w+') as f:
            f.write(html)

    def _generate_index(self):
        template = self.jinja_env.get_template(self.index_template)
        html = template.render(tests_sum=self._tests_sum)
        with codecs.open(os.path.join(self.dest, 'index.html'), encoding='utf-8', mode='w+') as f:
            f.write(html)


class XUnitReporter(Reporter):
    def __init__(self, dest='results',
                 summary_title='XUnit Reports',
                 templates_path=None,
                 report_template=None,
                 summary_template=None):
        super(XUnitReporter, self).__init__()
        self.dest = dest
        self._tests_sum = TestsSummary(summary_title)
        if templates_path and report_template and summary_template:
            loader = jinja2.FileSystemLoader(templates_path, encoding='utf-8')
            self.report_template = report_template
            self.summary_template = summary_template
        else:
            loader = jinja2.PackageLoader(unishark.PACKAGE, package_path='templates', encoding='utf-8')
            self.report_template = 'junit_suite_result.xml'
            self.summary_template = 'junit_suites_result.xml'
        self.jinja_env = jinja2.Environment(loader=loader, autoescape=False)

    def _make_output_dir(self):
        with Reporter.lock:
            if not os.path.exists(self.dest):
                os.makedirs(self.dest)

    def report(self, result):
        self._make_output_dir()
        # Summary
        suite_sum = SuiteSummary(result.name)
        suite_sum.build(result)
        # Get short method names
        for mod_sum in suite_sum.mod_sum_list:
            for cls_sum in mod_sum.cls_sum_list:
                for mth_sum in cls_sum.mth_sum_list:
                    mth_sum.name = mth_sum.name.split('.')[-1]
        # Generate report from single suite junit result template
        template = self.jinja_env.get_template(self.report_template)
        xml = template.render(suite_sum=suite_sum)
        filename = os.path.join(self.dest, suite_sum.name + '_xunit_result.xml')
        with codecs.open(filename, encoding='utf-8', mode='w+') as f:
            f.write(xml)

        self._tests_sum.suite_sum_list.append(suite_sum)

    def collect(self):
        self._make_output_dir()
        # Summary
        self._tests_sum.build(actual_duration=self._actual_duration)
        # Generate test suites summary
        template = self.jinja_env.get_template(self.summary_template)
        xml = template.render(tests_sum=self._tests_sum)
        with codecs.open(os.path.join(self.dest, 'summary_xunit_result.xml'), encoding='utf-8', mode='w+') as f:
            f.write(xml)


# Copyright 2015 Twitter, Inc and other contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from unittest.suite import TestSuite as UnitTestSuite, _ErrorHolder
from unittest.case import SkipTest
import sys
from unishark.util import get_module_name
from unishark.result import combine_results
import concurrent.futures
import logging

log = logging.getLogger(__name__)


def _is_suite(test):
    try:
        iter(test)
    except TypeError:
        return False
    return True


def _get_level(test):
    if not _is_suite(test):
        return TestSuite.METHOD_LEVEL
    else:
        for t in test:
            return _get_level(t) - 1
        return TestSuite.ROOT_LEVEL


def _get_current_module(test):
    # test is a module level suite
    current_module = None
    for tt in test:  # tt is a class level suite
        for t in tt:  # t is a test case instance
            current_module = t.__class__.__module__
            break
        break
    return current_module


def _get_current_class(test):
    # test is a class level suite
    current_class = None
    for t in test:  # t is a test case instance
        current_class = t.__class__
        break
    return current_class


def _call_if_exists(parent, attr):
    func = getattr(parent, attr, lambda: None)
    func()


def _group_test_cases(test, dic):
    if not _is_suite(test):
        mod = get_module_name(test)
        if mod not in dic:
            dic[mod] = dict()
        cls = test.__class__
        if cls not in dic[mod]:
            dic[mod][cls] = []
        dic[mod][cls].append(test)
    else:
        for t in test:
            _group_test_cases(t, dic)


def convert(test):
    suite = TestSuite()
    dic = dict()
    _group_test_cases(test, dic)
    for mod_dic in dic.values():
        mod_suite = TestSuite()
        for cases in mod_dic.values():
            cls_suite = TestSuite()
            cls_suite.addTests(cases)
            mod_suite.addTest(cls_suite)
        suite.addTest(mod_suite)
    log.debug('Converted tests: %r' % suite)
    return suite


class TestSuite(UnitTestSuite):
    ROOT_LEVEL = 0
    MODULE_LEVEL = 1
    CLASS_LEVEL = 2
    METHOD_LEVEL = 3

    def __init__(self, tests=()):
        super(TestSuite, self).__init__(tests)
        self._successful_fixtures = set()
        self._failed_fixtures = set()

    def __len__(self):
        return len(self._tests)

    def validate(self):
        assert len(self) > 0
        for mod_suite in self:
            assert _is_suite(mod_suite)
            assert isinstance(mod_suite, TestSuite)
            assert len(mod_suite) > 0
            for cls_suite in mod_suite:
                assert _is_suite(cls_suite)
                assert isinstance(cls_suite, TestSuite)
                assert len(cls_suite) > 0
                for case in cls_suite:
                    assert not _is_suite(case)

    def validate_result(self, result):
        assert len(self) == len(result)
        for mod_suite, mod_result in zip(self, result):
            assert len(mod_suite) == len(mod_result)
            assert type(mod_result) is type(result)
            for cls_suite, cls_result in zip(mod_suite, mod_result):
                assert len(cls_suite) == len(cls_result)
                assert type(cls_result) is type(result)
                for case_result in cls_result:
                    assert type(case_result) is type(result)

    def run(self, result, debug=False, concurrency_level=ROOT_LEVEL, max_workers=1, timeout=None):
        if concurrency_level < TestSuite.ROOT_LEVEL or concurrency_level > TestSuite.METHOD_LEVEL:
            raise ValueError('concurrency_level must be between %d and %d.'
                             % (TestSuite.ROOT_LEVEL, TestSuite.METHOD_LEVEL))
        if debug or self.countTestCases() <= 0 or max_workers <= 1:
            return super(TestSuite, self).run(result, debug=debug)
        self.validate()
        self.validate_result(result)
        with concurrent.futures.ThreadPoolExecutor(max_workers) as main_executor:
            self._run(self, result, TestSuite.ROOT_LEVEL, concurrency_level, main_executor, timeout)
        return result

    def _run(self, test, result, current_level, concurrency_level, main_executor, timeout):
        # test is a test suite instance which must be well-formed.
        # A well-formed test suite has a 4-level self-embedded structure:
        #                                    suite obj(root)
        #                                     /           \
        #                             suite obj(mod1)    suite obj(mod2)
        #                                /      \                    \
        #               suite obj(mod1.cls1)  suite obj(mod1.cls2)  ...
        #                 /            \                       \
        # case obj(mod1.cls1.mth1)  case obj(mod1.cls1.mth2)  ...
        if current_level == concurrency_level:
            self._seq_run(test, result)
        elif current_level < concurrency_level:
            results = result.children  # Divide: for each sub-suite(or test case) in the suite, there is a child result
            if current_level == TestSuite.ROOT_LEVEL:
                if concurrency_level == TestSuite.MODULE_LEVEL:
                    self._handle_fixtures(main_executor, self._setup_module, self._teardown_module,
                                          test, results, current_level, concurrency_level, main_executor, timeout)
                else:
                    with concurrent.futures.ThreadPoolExecutor(len(test)) as executor:
                        self._handle_fixtures(executor, self._setup_module, self._teardown_module,
                                              test, results, current_level, concurrency_level, main_executor, timeout)
            elif current_level == TestSuite.MODULE_LEVEL:
                if concurrency_level == TestSuite.CLASS_LEVEL:
                    self._handle_fixtures(main_executor, self._setup_class, self._teardown_class,
                                          test, results, current_level, concurrency_level, main_executor, timeout)
                else:
                    with concurrent.futures.ThreadPoolExecutor(len(test)) as executor:
                        self._handle_fixtures(executor, self._setup_class, self._teardown_class,
                                              test, results, current_level, concurrency_level, main_executor, timeout)
            elif current_level == TestSuite.CLASS_LEVEL:
                futures = [
                    main_executor.submit(self._run, t, r, current_level+1, concurrency_level, main_executor, timeout)
                    for t, r in zip(test, results)
                ]
                concurrent.futures.wait(futures, timeout=timeout)
            combine_results(result, results)  # Conquer: collect child results into parent result
        return test, result

    def _seq_run(self, test, result):
        level = _get_level(test)
        if level == TestSuite.ROOT_LEVEL:
            for t in test:
                self._seq_run(t, result)
        elif level == TestSuite.MODULE_LEVEL:
            self._setup_module(test, result)
            for t in test:
                self._seq_run(t, result)
            self._teardown_module(test, result)
        elif level == TestSuite.CLASS_LEVEL:
            self._setup_class(test, result)
            for t in test:
                self._seq_run(t, result)
            self._teardown_class(test, result)
        elif level == TestSuite.METHOD_LEVEL:
            current_class = test.__class__
            current_module = current_class.__module__
            if current_module + '.setUpModule' in self._failed_fixtures:
                return
            if '.'.join((current_module, current_class.__name__, 'setUpClass')) in self._failed_fixtures:
                return
            test(result)
        else:
            raise NotImplementedError

    def _handle_fixtures(self, executor, setup_fn, teardown_fn,
                         test, results, current_level, concurrency_level, main_executor, timeout):
        futures_of_setup = [executor.submit(setup_fn, t, r) for t, r in zip(test, results)]
        futures = []
        for done in concurrent.futures.as_completed(futures_of_setup, timeout=timeout):
            t, r = done.result()
            futures.append(
                executor.submit(self._run, t, r, current_level+1, concurrency_level, main_executor, timeout)
            )
        futures_of_teardown = []
        for done in concurrent.futures.as_completed(futures, timeout=timeout):
            t, r = done.result()
            futures_of_teardown.append(executor.submit(teardown_fn, t, r))
        concurrent.futures.wait(futures_of_teardown, timeout=timeout)

    def _addClassOrModuleLevelException(self, result, exception, error_name):
        error = FixtureErrors(error_name)
        addSkip = getattr(result, 'addSkip', None)
        if addSkip is not None and isinstance(exception, SkipTest):
            addSkip(error, str(exception))
        else:
            result.addError(error, sys.exc_info())

    def _setup_module(self, test, result):
        # test must be a module level suite
        current_module = _get_current_module(test)
        fixture_name = current_module + '.setUpModule'
        if fixture_name in self._successful_fixtures or fixture_name in self._failed_fixtures:
            return test, result

        try:
            module = sys.modules[current_module]
        except KeyError:
            return test, result
        setUpModule = getattr(module, 'setUpModule', None)
        if setUpModule is not None:
            _call_if_exists(result, '_setupStdout')
            try:
                setUpModule()
                self._successful_fixtures.add(fixture_name)
            except Exception as e:
                self._failed_fixtures.add(fixture_name)
                error_name = '%s:setUpModule' % current_module
                self._addClassOrModuleLevelException(result, e, error_name)
            finally:
                _call_if_exists(result, '_restoreStdout')
        return test, result

    def _teardown_module(self, test, result):
        # test must be a module level suite
        current_module = _get_current_module(test)
        fixture_name = current_module + '.tearDownModule'
        if fixture_name in self._successful_fixtures or fixture_name in self._failed_fixtures:
            return
        if current_module + '.setUpModule' in self._failed_fixtures:
            return

        try:
            module = sys.modules[current_module]
        except KeyError:
            return
        tearDownModule = getattr(module, 'tearDownModule', None)
        if tearDownModule is not None:
            _call_if_exists(result, '_setupStdout')
            try:
                tearDownModule()
                self._successful_fixtures.add(fixture_name)
            except Exception as e:
                self._failed_fixtures.add(fixture_name)
                error_name = '%s:tearDownModule' % current_module
                self._addClassOrModuleLevelException(result, e, error_name)
            finally:
                _call_if_exists(result, '_restoreStdout')

    def _setup_class(self, test, result):
        # test must be a class level suite
        current_class = _get_current_class(test)
        current_module = current_class.__module__
        class_name = '.'.join((current_module, current_class.__name__))
        fixture_name = class_name + '.setUpClass'
        if fixture_name in self._successful_fixtures or fixture_name in self._failed_fixtures:
            return test, result
        if current_module + '.setUpModule' in self._failed_fixtures:
            return test, result
        if getattr(current_class, '__unittest_skip__', False):
            return test, result

        setUpClass = getattr(current_class, 'setUpClass', None)
        if setUpClass is not None:
            _call_if_exists(result, '_setupStdout')
            try:
                setUpClass()
                self._successful_fixtures.add(fixture_name)
            except Exception as e:
                self._failed_fixtures.add(fixture_name)
                current_class._classSetupFailed = True
                error_name = '%s:%s:setUpClass' % (current_module, current_class.__name__)
                self._addClassOrModuleLevelException(result, e, error_name)
            finally:
                _call_if_exists(result, '_restoreStdout')
        return test, result

    def _teardown_class(self, test, result):
        # test must be a class level suite
        current_class = _get_current_class(test)
        current_module = current_class.__module__
        class_name = '.'.join((current_module, current_class.__name__))
        fixture_name = class_name + '.tearDownClass'
        if fixture_name in self._successful_fixtures or fixture_name in self._failed_fixtures:
            return
        if class_name + '.setUpClass' in self._failed_fixtures:
            return
        if current_module + '.setUpModule' in self._failed_fixtures:
            return
        if getattr(current_class, "__unittest_skip__", False):
            return

        tearDownClass = getattr(current_class, 'tearDownClass', None)
        if tearDownClass is not None:
            _call_if_exists(result, '_setupStdout')
            try:
                tearDownClass()
                self._successful_fixtures.add(fixture_name)
            except Exception as e:
                self._failed_fixtures.add(fixture_name)
                error_name = '%s:%s:tearDownClass' % (current_module, current_class.__name__)
                self._addClassOrModuleLevelException(result, e, error_name)
            finally:
                _call_if_exists(result, '_restoreStdout')


class FixtureErrors(_ErrorHolder):
    def __init__(self, description):
        super(FixtureErrors, self).__init__(description)
        self.description = self.description.replace('.', ':')

    def id(self):
        return "%s.%s" % (self.__class__.__name__, self.description)


from setuptools import setup, find_packages
from os import path
from sys import version_info

NAME = 'unishark'
VERSION = '0.3.1'

long_description = (
    open(path.join('docs', 'README.rst'), 'r').read() + '\n' +
    open(path.join('docs', 'CHANGELOG.rst'), 'r').read()
)

py3_requires = ['Jinja2', 'MarkupSafe']
py2_requires = ['Jinja2', 'MarkupSafe', 'futures']

requires = None
if version_info[0] >= 3 and version_info[1] >= 2:
    requires = py3_requires
else:
    requires = py2_requires

setup(
    name=NAME,
    version=VERSION,

    description='A test framework extending unittest, providing flexible test suites config, concurrent execution, '
                'Html/XUnit reports, and data driven utility.',
    long_description=long_description,

    url='https://github.com/twitter/unishark',

    author='Ying Ni',
    author_email='niying7@gmail.com',

    license='Apache License, Version 2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',

        'License :: OSI Approved :: Apache Software License',

        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',

        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    keywords='test framework, unittest extension, concurrent, data driven',

    # See https://packaging.python.org/en/latest/requirements.html
    install_requires=requires,

    extras_require={
        'test': ['coveralls'],
    },

    packages=find_packages(include=[NAME]),
    include_package_data=True,
    package_data={
        NAME: ['templates/*.html', 'templates/*.xml'],
    },
)
